using System;
using System.IO;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using FitConnect;
using FluentAssertions;
using Microsoft.Extensions.Logging;
using NUnit.Framework;

namespace SenderTest;

public class SenderEncryptionWithSelfSignedCertificateTest {
    private Sender _sender = null!;
    private ILogger<SenderEncryptionWithSelfSignedCertificateTest> _logger = null!;
    private X509Certificate2 _certificate = null!;
    private static byte[]? cypher;

    private const string ToEncrypt = "This is a test message";


    /*
     * Encryption test must be changed for production to only allow extern signed certificates
     * and forbid self-signed certificates.
     */

    [OneTimeSetUp]
    public void OneTimeSetup() {
        _certificate = CreateSelfSignedCertificate("./");
    }

    [OneTimeTearDown]
    public void OneTimeTearDown() {
        _certificate.Dispose();
        File.Delete("./certificate.pfx");
    }

    [SetUp]
    public void Setup() {
        _logger = LoggerFactory.Create(cfg => cfg.AddConsole())
            .CreateLogger<SenderEncryptionWithSelfSignedCertificateTest>();
        _sender = new Sender(_logger,
            FitConnectEndpoints.Create(FitConnectEndpoints.EndpointType.Development));

        var certificate = new X509Certificate2("./certificate.pfx");
        _sender.ImportCertificate(certificate);
    }


    [Test]
    [Order(10)]
    public void CryptWithOutPublicKeyImport() {
        cypher = _sender.EncryptData(Encoding.UTF8.GetBytes(ToEncrypt));

        _logger.LogInformation("Cypher: {}", Convert.ToBase64String(cypher));
    }

    [Test]
    [Order(20)]
    public void Decrypt_ResultShouldMatchToEncrypt() {
        var result = _sender.DecryptDataAsync(cypher!);
        Encoding.UTF8.GetString(result).Should().Be(ToEncrypt);
    }


    [Test]
    public void ExportPrivateKey() {
        var privateKey = _sender.ExportPrivateKey();
        _logger.LogInformation("Private key: {}", Convert.ToBase64String(privateKey));
    }


    #region Static helpers - Certificates

    /// <summary>
    /// 
    /// </summary>
    /// <param name="exportPath">The path to export the certificate.
    /// <para>"../../../" matches the development path of the project</para></param>
    /// <returns></returns>
    /// <exception cref="Exception"></exception>
    private X509Certificate2 CreateSelfSignedCertificate(string? exportPath = null) {
        var rsa = RSA.Create(4096);

        var req = new CertificateRequest("c=DE, cn=fitconnect.de",
            rsa, HashAlgorithmName.SHA256, RSASignaturePadding.Pkcs1);

        var cert = req.CreateSelfSigned(DateTimeOffset.Now.AddSeconds(-5),
            DateTimeOffset.Now.AddYears(5));

        if (cert.GetRSAPublicKey() == null)
            throw new Exception("Certificate does not contain a public key");

        if (cert.GetRSAPrivateKey() == null)
            throw new Exception("Certificate does not contain a private key");

        // Export the certificate to a PEM file, just for
        // additional extern testing
        if (exportPath != null) {
            ExportCertificateToFile(exportPath, cert);
        }

        return cert;
    }

    private void ExportCertificateToFile(string exportPath, X509Certificate cert) {
        // Create PFX (PKCS #12) with private key
        File.WriteAllBytes($"{exportPath}/certificate.pfx",
            cert.Export(X509ContentType.Pfx, ""));

        // Create Base 64 encoded CER (public key only)
        File.WriteAllText($"{exportPath}/certificate.cer",
            "-----BEGIN CERTIFICATE-----\r\n"
            + Convert.ToBase64String(cert.Export(X509ContentType.Cert, ""),
                Base64FormattingOptions.InsertLineBreaks)
            + "\r\n-----END CERTIFICATE-----");

        _logger?.LogInformation("Exporting {}",
            Convert.ToBase64String(cert.Export(X509ContentType.Cert, "")));
    }

    #endregion
}
